configs = [
    {"id": "/root/test1/id1", "value": "value1", "label": None},
    {"id": "/root/test1/id2", "value": "value2", "label": None},
    {"id": "/root/test2/id1", "value": "value3", "label": None},
    {"id": "/root/test2/id2", "value": "value4", "label": None},
    {"id": "s3", "value": "value5", "label": None},
    {"id": "s4", "value": "value6", "label": None},
    {"id": "/root/test1/id1", "value": "value7", "label": "label1"},
    {"id": "/root/test2/id2", "value": "value8", "label": "label1"},
    {"id": "s3", "value": "value11", "label": "label1"},
    {"id": "/root/test1/id2", "value": "value9", "label": "label2"},
    {"id": "/root/test2/id1", "value": "value10", "label": "label2"},
    {"id": "s5", "value": "value12", "label": "label2"},
]
