documents = [
    {
        "id": "id00",
        "pk": "pk00",
        "str": "zero",
        "int": 0,
        "float": 0.0,
        "bool": True,
        "empty": None,
        "const": "#",
        "text1": "The quick brown fox jumps over the lazy yellow dog.",
        "text2": "Zero-based indexing starts at index 0 in most languages.",
        "vector": [-1, -1, -1, -1],
        "sparse_vector": {
            "0": 1,
            "2": 1,
        },
        "obj": {
            "nstr": "0",
            "nint": -0,
            "nopt": 0,
            "narr": None,
            "nobj": {"nnstr": "", "nnfloat": -0.0},
        },
        "arrint": [0, 0, 0],
        "arrstr": ["ten", "hundred", "thousand", "zero"],
        "arrobj": [
            {"ostr": "million", "oint": 0},
            {"ostr": "billion", "oint": 0},
            {"ostr": "trillion", "oint": 0},
        ],
    },
    {
        "id": "id01",
        "pk": "pk00",
        "str": "one",
        "int": 1,
        "float": 0.1,
        "bool": False,
        "empty": None,
        "const": "#",
        "text1": "One small step for man, one giant leap for mankind.",
        "text2": "First principles help solve complex problems.",
        "vector": [1, 1, 1, 1],
        "sparse_vector": {
            "1": 2,
            "3": 2,
        },
        "obj": {
            "nstr": "1",
            "nint": -1,
            "nopt": 1,
            "narr": [],
            "nobj": {"nnstr": "i", "nnfloat": -100.100},
        },
        "arrint": [11, 101, 1001],
        "arrstr": ["eleven", "hundred one", "thousand one"],
        "arrobj": [
            {"ostr": "one million", "oint": 1000000},
            {"ostr": "one billion", "oint": 1000000000},
            {"ostr": "one trillion", "oint": 100000000000},
        ],
    },
    {
        "id": "id02",
        "pk": "pk00",
        "str": "two",
        "int": 2,
        "float": 0.2,
        "bool": True,
        "empty": None,
        "opt": "abcd",
        "const": "#",
        "text1": "Two roads diverged in a wood.",
        "text2": "Pair programming improves yellow code quality.",
        "vector": [2, 2, 2, 2],
        "sparse_vector": {
            "0": 3,
            "2": 3,
        },
        "obj": {
            "nstr": "2",
            "nint": -2,
            "nopt": 2,
            "narr": [200],
            "nobj": {"nnstr": "ii", "nnfloat": -200.200},
        },
        "arrint": [22, 202, 2002],
        "arrstr": ["twelve", "hundred two", "thousand two", "zero"],
        "arrobj": [
            {"ostr": "two million", "oint": 2000000},
            {"ostr": "two billion", "oint": 2000000000},
            {"ostr": "two trillion", "oint": 200000000000},
        ],
    },
    {
        "id": "id03",
        "pk": "pk00",
        "str": "three",
        "int": 3,
        "float": 0.3,
        "bool": False,
        "null": None,
        "opt": "bcde",
        "const": "#",
        "text1": "Three is a crowd but also a charm in stories.",
        "text2": "Triangles have three sides and three angles.",
        "vector": [3, 3, 3, 3],
        "sparse_vector": {
            "1": 4,
            "3": 4,
        },
        "obj": {
            "nstr": "3",
            "nint": -3,
            "nopt": 3,
            "narr": [300, 310],
            "nobj": {"nnstr": "iii", "nnfloat": -300.300},
        },
        "arrint": [33, 303, 3003],
        "arrstr": ["thirteen", "hundred three", "thousand three", "zero"],
        "arrobj": [
            {"ostr": "three million", "oint": 3000000},
            {"ostr": "three billion", "oint": 3000000000},
            {"ostr": "three trillion", "oint": 300000000000},
        ],
    },
    {
        "id": "id04",
        "pk": "pk00",
        "str": "four",
        "int": 4,
        "float": 0.4,
        "bool": True,
        "empty": None,
        "opt": "cdef",
        "const": "#",
        "text1": "Four score and seven years ago began a famous speech.",
        "text2": "Quarter finals are the fourth round in many tournaments.",
        "vector": [4, 4, 4, 4],
        "sparse_vector": {
            "0": 5,
            "4": 5,
        },
        "obj": {
            "nstr": "4",
            "nint": -4,
            "nopt": 4,
            "narr": [400, 410, 420],
            "nobj": {"nnstr": "iv", "nnfloat": -400.400},
        },
        "arrint": [44, 404, 4004],
        "arrstr": ["fourteen", "hundred four", "thousand four", "zero"],
        "arrobj": [
            {"ostr": "four million", "oint": 4000000},
            {"ostr": "four billion", "oint": 4000000000},
            {"ostr": "four trillion", "oint": 400000000000},
        ],
    },
    {
        "id": "id05",
        "pk": "pk01",
        "str": "five",
        "int": 5,
        "float": 0.5,
        "bool": False,
        "empty": None,
        "opt": "defg",
        "const": "#",
        "text1": "Five golden rings were part of the carol.",
        "text2": "High five celebrates success and teamwork.",
        "vector": [5, 5, 5, 5],
        "sparse_vector": {
            "1": 6,
            "4": 6,
        },
        "obj": {
            "nstr": "5",
            "nint": -5,
            "nopt": 5,
            "narr": [500, 510, 520, 530],
            "nobj": {"nnstr": "v", "nnfloat": -500.500},
        },
        "arrint": [55, 505, 5005],
        "arrstr": ["fifteen", "hundred five", "thousand five"],
        "arrobj": [
            {"ostr": "five million", "oint": 5000000},
            {"ostr": "five billion", "oint": 5000000000},
            {"ostr": "five trillion", "oint": 500000000000},
        ],
    },
    {
        "id": "id06",
        "pk": "pk01",
        "str": "xsix",
        "int": 6,
        "float": 0.6,
        "bool": True,
        "empty": None,
        "opt": "efgh",
        "const": "#",
        "text1": "Six degrees of separation connect the world.",
        "text2": "The hexagon has six equal sides.",
        "vector": [6, 6, 6, 6],
        "sparse_vector": {
            "0": 6,
            "5": 6,
        },
        "obj": {
            "nstr": "6",
            "nint": -6,
            "narr": [600, 610, 620, 630, 640],
            "nobj": {"nnstr": "vi", "nnfloat": -600.600},
        },
        "arrint": [66, 606, 6006],
        "arrstr": ["sixteen", "hundred six", "thousand six"],
        "arrobj": [
            {"ostr": "six million", "oint": 6000000},
            {"ostr": "six billion", "oint": 6000000000},
            {"ostr": "six trillion", "oint": 600000000000},
        ],
    },
    {
        "id": "id07",
        "pk": "pk01",
        "str": "xseven",
        "int": 7,
        "float": 0.7,
        "bool": False,
        "empty": None,
        "opt": "fghi",
        "const": "#",
        "text1": "Seven continents span our planet.",
        "text2": "Lucky number seven appears in myths.",
        "vector": [7, 7, 7, 7],
        "sparse_vector": {
            "2": 1,
            "5": 1,
        },
        "obj": {
            "nstr": "7",
            "nint": -7,
            "narr": [700, 710, 720, 730, 740, 750],
            "nobj": {"nnstr": "vii", "nnfloat": -700.700},
        },
        "arrint": [77, 707, 7007],
        "arrstr": ["seventeen", "hundred seven", "thousand seven"],
        "arrobj": [
            {"ostr": "seven million", "oint": 7000000},
            {"ostr": "seven billion", "oint": 7000000000},
            {"ostr": "seven trillion", "oint": 700000000000},
        ],
    },
    {
        "id": "id08",
        "pk": "pk01",
        "str": "eight",
        "int": 8,
        "float": 0.8,
        "bool": True,
        "empty": None,
        "const": "#",
        "text1": "Eight bits make a byte in computing.",
        "text2": "Octopuses have eight arms and remarkable cats intelligence.",
        "vector": [8, 8, 8, 8],
        "sparse_vector": {
            "3": 2,
            "6": 2,
        },
        "obj": {
            "nstr": "8",
            "nint": -8,
            "nopt": None,
            "narr": [800, 810, 820, 830, 840, 850, 860],
            "nobj": {"nnstr": "viii", "nnfloat": -800.800},
        },
        "arrint": [88, 808, 8008],
        "arrstr": ["eighteen", "hundred eight", "thousand eight"],
        "arrobj": [
            {"ostr": "eight million", "oint": 8000000},
            {"ostr": "eight billion", "oint": 8000000000},
            {"ostr": "eight trillion", "oint": 800000000000},
        ],
    },
    {
        "id": "id09",
        "pk": "pk01",
        "str": "nine",
        "int": 9,
        "float": 0.9,
        "bool": False,
        "empty": None,
        "const": "#",
        "hard": "doesn't",
        "text1": "Nine lives are attributed to cats in folklore.",
        "text2": "Cloud nine describes a state of bliss.",
        "vector": [9, 9, 9, 9],
        "sparse_vector": {
            "4": 9,
            "5": 9,
        },
        "obj": {
            "nstr": "9",
            "nint": -9,
            "nopt": None,
            "narr": [900, 910, 920, 930, 940, 950, 960, 970],
            "nobj": {"nnstr": "ix", "nnfloat": -900.900},
        },
        "arrint": [99, 909, 9009],
        "arrstr": ["nineteen", "hundred nine", "thousand nine"],
        "arrobj": [
            {"ostr": "nine million", "oint": 9000000},
            {"ostr": "nine billion", "oint": 9000000000},
            {"ostr": "nine trillion", "oint": 900000000000},
        ],
    },
]
