kvs = [
    {
        "id": "id00",
        "value": b"Hello World 1",
    },
    {
        "id": "id01",
        "value": 10,
    },
    {
        "id": "id02",
        "value": "Hello World ",
    },
]

batch_kvs = [
    {
        "id": "id00",
        "value": b"Batch 0",
        "collection": "col1",
    },
    {
        "id": "id01",
        "value": b"Batch 1",
        "collection": "col2",
    },
    {
        "id": "id02",
        "value": b"Batch 2",
        "collection": "col1",
    },
    {
        "id": "id03",
        "value": b"Batch 3",
        "collection": "col2",
    },
]

query_kvs = [
    {
        "id": "q00",
        "value": b"value0",
        "collection": "col1",
    },
    {
        "id": "q01",
        "value": b"value1",
        "collection": "col1",
    },
    {
        "id": "q02",
        "value": b"value2",
        "collection": "col1",
    },
    {
        "id": "q03",
        "value": b"value3",
        "collection": "col1",
    },
    {
        "id": "q04",
        "value": b"value4",
        "collection": "col1",
    },
    {
        "id": "q05",
        "value": b"value5",
        "collection": "col2",
    },
    {
        "id": "q06",
        "value": b"value6",
        "collection": "col2",
    },
    {
        "id": "q07",
        "value": b"value7",
        "collection": "col2",
    },
    {
        "id": "q08",
        "value": b"value8",
        "collection": "col2",
    },
    {
        "id": "q09",
        "value": b"value9",
        "collection": "col2",
    },
    {
        "id": "q10",
        "value": b"value10",
        "collection": "col2",
    },
    {
        "id": "q11",
        "value": b"value11",
        "collection": "col2",
    },
    {
        "id": "q12",
        "value": b"value12",
        "collection": "col2",
    },
    {
        "id": "q13",
        "value": b"value13",
        "collection": "col2",
    },
]
