class MessagingOperation:
    CREATE_QUEUE = "create_queue"
    UPDATE_QUEUE = "update_queue"
    DROP_QUEUE = "drop_queue"
    LIST_QUEUES = "list_queues"
    HAS_QUEUE = "has_queue"
    GET_QUEUE = "get_queue"

    CREATE_TOPIC = "create_topic"
    UPDATE_TOPIC = "update_topic"
    DROP_TOPIC = "drop_topic"
    LIST_TOPICS = "list_topics"
    HAS_TOPIC = "has_topic"
    GET_TOPIC = "get_topic"

    CREATE_SUBSCRIPTION = "create_subscription"
    UPDATE_SUBSCRIPTION = "update_subscription"
    DROP_SUBSCRIPTION = "drop_subscription"
    LIST_SUBSCRIPTIONS = "list_subscriptions"
    HAS_SUBSCRIPTION = "has_subscription"
    GET_SUBSCRIPTION = "get_subscription"

    PUT = "put"
    BATCH = "batch"
    PULL = "pull"
    ACK = "ack"
    NACK = "nack"
    EXTEND = "extend"
    PURGE = "purge"
    CLOSE = "close"
