secrets = [
    {
        "id": "test-id-1",
        "values": [
            "test-1-value-1",
            "test-1-value-2",
            "test-1-value-3",
            "test-1-value-4",
        ],
        "metadata": [
            {
                "test-1-tag-1-key-1": "test-1-tag-1-value-1",
                "test-1-tag-1-key-2": "test-1-tag-1-value-2",
                "common-key": "common-value",
            },
            {
                "test-1-tag-2-key-1": "test-1-tag-2-value-1",
                "test-1-tag-2-key-2": "test-1-tag-2-value-2",
            },
        ],
    },
    {
        "id": "test-id-2",
        "values": ["test-2-value-2"],
        "metadata": [
            {
                "test-2-tag-1-key-1": "test-2-tag-1-value-1",
                "test-2-tag-1-key-2": "test-2-tag-1-value-2",
                "common-key": "common-value",
            }
        ],
    },
    {
        "id": "test-id-11",
        "values": ["test-11-value-2"],
        "metadata": [
            {
                "test-11-tag-1-key-1": "test-11-tag-1-value-1",
                "test-11-tag-1-key-2": "test-11-tag-1-value-2",
            }
        ],
    },
]
